import json
from importlib.resources import read_text

from scrapy.spiders import CrawlSpider

from products_crawler import spider_configs
from products_crawler import items, itemloaders


class GenericSpider(CrawlSpider):
    def __init__(self, *args, **kwargs):
        super(GenericSpider, self).__init__()

        self.start_urls = kwargs.pop('start_urls', list())
        self.logger.debug('original start_urls is %r' % self.start_urls)
        if isinstance(self.start_urls, str):
            try:
                self.start_urls = json.loads(self.start_urls)
            except json.JSONDecodeError:
                self.start_urls = [self.start_urls]
        self.logger.debug('cleaned start_urls is %r' % self.start_urls)

        config = self._get_spider_cfg()
        self.item_class = getattr(items, config['item_class'])
        self.loader_class = getattr(itemloaders, config['loader_class'])
        self.load_rules = config.get('load_rules', {})

    def _get_spider_cfg(self):
        content = read_text(spider_configs, f'{self.name}.json')
        return json.loads(content)

    def parse(self, response, **kwargs):
        item = self.item_class()
        loader = self.loader_class(item=item, response=response)

        for field, rule in self.load_rules.items():
            if rule["method"] == "":
                continue
            if rule["method"] == "value":
                loader.add_value(field, *rule["args"])
            elif rule["method"] == "xpath":
                loader.add_xpath(field, *rule["args"])
            elif rule["method"] == "css":
                loader.add_css(field, *rule["args"])
            elif rule["method"] == "eval":
                loader.add_value(field, *map(eval, rule["args"]))
            else:
                raise ValueError("No such method: {0}".format(rule["method"]))

        return loader.load_item()

    def parse_start_url(self, response, **kwargs):
        return self.parse(response)



# -*- coding: utf-8 -*-
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Rule
from . import GenericSpider


class GlldSpider(GenericSpider):
    name = 'glld'
    allowed_domains = ['gallianolandor.com']
    rules = (
        # Match each item in item list.
        Rule(
            LinkExtractor(
                allow=r'.*?products.*?',
                restrict_xpaths=r'//div[@data-section-id="collection-template"]'
            ),
            callback='parse',
        ),
        # Match other pages.
        Rule(
            LinkExtractor(
                restrict_xpaths=r'//ul[@class="pagination-custom"]'
            )
        )
    )



# -*- coding: utf-8 -*-
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Rule
from . import GenericSpider


class NikeSpider(GenericSpider):
    name = 'nike'
    allowed_domains = ['nike.com']
    custom_settings = dict(
            DOWNLOADER_MIDDLEWARES={
                'products_crawler.middlewares.SeleniumDownloaderMiddleware': 543,
            }
    )

    rules = (
        # Match each product in product list.
        Rule(
            LinkExtractor(restrict_xpaths=r'//div[contains(@class, "product-grid__items")]', deny=r'.+jpg'),
            callback='parse',
            follow=True
        ),
        # Match products of other color in product detail.
        Rule(
            LinkExtractor(restrict_xpaths=r'//div[@class="colorway-images-wrapper"]', deny=r'.+jpg'),
            callback='parse',
        )
    )



import os
from argparse import ArgumentParser

from scrapy.crawler import CrawlerProcess
from scrapy.utils.project import get_project_settings
from scrapy.spiderloader import SpiderLoader


def main():
    os.environ.setdefault('SCRAPY_SETTINGS_MODULE', 'products_crawler.settings')
    scrapy_settings = get_project_settings()
    spider_loader = SpiderLoader(scrapy_settings)

    parse = ArgumentParser()
    parse.add_argument("brand", choices=spider_loader.list())
    parse.add_argument("start_urls", nargs="+")
    args = parse.parse_args()

    process = CrawlerProcess(scrapy_settings)
    process.crawl(args.brand, **{'start_urls': args.start_urls})
    process.start()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Scrapy settings for products_crawler project
#
# For simplicity, this file contains only settings considered important or
# commonly used. You can find more settings consulting the documentation:
#
#     https://docs.scrapy.org/en/latest/topics/settings.html
#     https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
#     https://docs.scrapy.org/en/latest/topics/spider-middleware.html

BOT_NAME = 'products_crawler'

SPIDER_MODULES = ['products_crawler.spiders']
NEWSPIDER_MODULE = 'products_crawler.spiders'

IMAGES_STORE = r'products'

SELENIUM_TIMEOUT = 20
CHROME_OPTIONS = ['--headless']

# Crawl responsibly by identifying yourself (and your website) on the user-agent
# USER_AGENT = 'products_crawler (+http://www.yourdomain.com)'

# Obey robots.txt rules
ROBOTSTXT_OBEY = True

# Configure maximum concurrent requests performed by Scrapy (default: 16)
#CONCURRENT_REQUESTS = 32

# Configure a delay for requests for the same website (default: 0)
# See https://docs.scrapy.org/en/latest/topics/settings.html#download-delay
# See also autothrottle settings and docs
DOWNLOAD_DELAY = 0.1
# The download delay setting will honor only one of:
#CONCURRENT_REQUESTS_PER_DOMAIN = 16
#CONCURRENT_REQUESTS_PER_IP = 16

# Disable cookies (enabled by default)
#COOKIES_ENABLED = False

# Disable Telnet Console (enabled by default)
#TELNETCONSOLE_ENABLED = False

# Override the default request headers:
#DEFAULT_REQUEST_HEADERS = {
#   'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
#   'Accept-Language': 'en',
#}

# Enable or disable spider middlewares
# See https://docs.scrapy.org/en/latest/topics/spider-middleware.html
#SPIDER_MIDDLEWARES = {
#    'products_crawler.middlewares.ProductCrawlerSpiderMiddleware': 543,
#}

# Enable or disable downloader middlewares
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
#DOWNLOADER_MIDDLEWARES = {
#    'products_crawler.middlewares.ProductCrawlerDownloaderMiddleware': 543,
#}

# Enable or disable extensions
# See https://docs.scrapy.org/en/latest/topics/extensions.html
#EXTENSIONS = {
#    'scrapy.extensions.telnet.TelnetConsole': None,
#}

# Configure item pipelines
# See https://docs.scrapy.org/en/latest/topics/item-pipeline.html
ITEM_PIPELINES = {
    'products_crawler.pipelines.ProductInfoPipeline': 300,
    'products_crawler.pipelines.ProductImagesPipeline': 301,
}

# Enable and configure the AutoThrottle extension (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/autothrottle.html
AUTOTHROTTLE_ENABLED = True
# The initial download delay
AUTOTHROTTLE_START_DELAY = 0.2
# The maximum download delay to be set in case of high latencies
AUTOTHROTTLE_MAX_DELAY = 0.5
# The average number of requests Scrapy should be sending in parallel to
# each remote server
AUTOTHROTTLE_TARGET_CONCURRENCY = 3.0
# Enable showing throttling stats for every response received:
# AUTOTHROTTLE_DEBUG = True

# Enable and configure HTTP caching (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html#httpcache-middleware-settings
#HTTPCACHE_ENABLED = True
#HTTPCACHE_EXPIRATION_SECS = 0
#HTTPCACHE_DIR = 'httpcache'
#HTTPCACHE_IGNORE_HTTP_CODES = []
#HTTPCACHE_STORAGE = 'scrapy.extensions.httpcache.FilesystemCacheStorage'

# MAIL_FROM = 'example@host'

# MAIL_HOST = 'smtp.host'

MAIL_PORT = 465

# MAIL_USER = 'example@host

# MAIL_PASS = ''

MAIL_TLS = True

MAIL_SSL = True

# STATSMAILER_RCPTS = []


EXTENSIONS = {
    'scrapy.extensions.statsmailer.StatsMailer': 500,
}



# -*- coding: utf-8 -*-
from scrapy.spiders import Rule
from scrapy.linkextractors import LinkExtractor
from . import GenericSpider


class KapitalSpider(GenericSpider):
    name = 'kapital'
    allowed_domains = ['kapital-webshop.jp']
    rules = [
        # Match each product in product list.
        Rule(
            LinkExtractor(
                allow=r'.*html',
                restrict_xpaths='//ul[@class="item_list clearfix"]/li//div[@class="img_inner"]/a'
            ),
            callback='parse',
        ),
        # Match other pages.
        Rule(
            LinkExtractor(
                restrict_xpaths='//div[@class="pager"]//a'
            ),
        )
    ]



import re

from itemloaders.processors import TakeFirst, Identity, Compose


take_first = TakeFirst()
identity = Identity()
take_first_and_strip = Compose(take_first, lambda x: x.strip())


def match_supreme_week(text):
    return re.findall(r"\((.*?)\)", text)[0]


def match_kapital_category(url):
    return re.findall(r'category/(.*?)/', url)[0]


def match_nike_art_no(text):
    return re.search(r'.{6}-.{3}', text).group()


def match_bearbrick_category(text):
    match = re.search(r'(\d{3}[％%] & )?\d{3,4}[％%]', text)
    if not match:
        return 'other'
    return match.group()



# -*- coding: utf-8 -*-
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Rule
from . import GenericSpider


class BearbrickSpider(GenericSpider):
    name = 'bearbrick'
    allowed_domains = ['bearbrick.com']
    rules = (
        # Match each product in product list.
        Rule(
            LinkExtractor(
                allow=r'.*html',
                restrict_xpaths='//section[@id="itemList"]'
            ), callback='parse'
        ),
        # Match other pages.
        Rule(
            LinkExtractor(
                restrict_xpaths='//div[@id="pager"]',
            ),
        )
    )



from scrapy.loader import ItemLoader
from itemloaders.processors import MapCompose, Compose

from products_crawler.processors import take_first, take_first_and_strip, identity, \
                                       match_supreme_week, match_kapital_category, \
                                       match_nike_art_no, match_bearbrick_category


class ProductLoader(ItemLoader):
    default_output_processor = take_first
    images_out = identity

    @classmethod
    def delete_url_query_arguments(cls, url):
        return url.rsplit('?', maxsplit=1)[0]


class SupremeLoader(ProductLoader):
    title_out = take_first_and_strip
    images_out = MapCompose(lambda x: x.replace("thumb", "sqr"))
    week_out = Compose(take_first, match_supreme_week)
    season_out = take_first_and_strip


class KapitalLoader(ProductLoader):
    art_no_out = take_first_and_strip
    category_out = Compose(take_first, match_kapital_category)


class GallianolandorLoader(ProductLoader):
    price_out = take_first_and_strip
    images_out = MapCompose(ProductLoader.delete_url_query_arguments)
    season_out = Compose(take_first, lambda x: x.replace('/', ''))


class NikeLoader(ProductLoader):
    images_out = Compose(
        identity,
        lambda urls: [url for url in urls if 'LOADING' not in url]
    )
    art_no_out = Compose(take_first, match_nike_art_no)


class BearBrickLoader(ProductLoader):
    title_out = take_first_and_strip
    category_in = Compose(take_first, match_bearbrick_category)


class UastoreLoader(ProductLoader):
    brand_out = Compose(take_first, lambda x: x.title())


class TsLoader(ProductLoader):
    price_out = take_first_and_strip
    images_out = MapCompose(ProductLoader.delete_url_query_arguments)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os.path
from urllib.parse import unquote
import json


ILLEGAL_FILENAME_CHARS = {"?", "/", "\\", ":", "*", ">", "<", '"', "'"}


def file_path(info, filename=''):
    """
    传入一个包含Item信息的字典，为商品选择一个保存的位置
    :param info: Request.meta or item object.
    :param filename: file name.
    :return: 相对于IMAGE_STORE的路径

    >>> info = {'brand': 'supreme', 'title': 'Boat Tee', 'art_no': '',\
     'season': '20SS', 'week': 'Week18', 'category': '',}

    >>> file_path(info)
    'supreme\\\\20SS\\\\Week18\\\\Boat Tee\\\\'
    >>> file_path(info, 'img.jpg')
    'supreme\\\\20SS\\\\Week18\\\\Boat Tee\\\\img.jpg'
    """
    brand = info.get('brand', '')
    title = info.get('title', '')
    art_no = info.get('art_no', '')
    season = info.get('season', '')
    week = info.get('week', '')
    category = info.get('category', '')
    if art_no and title:
        product = art_no + '-' + legal_name(title)
    elif art_no and not title:
        product = art_no
    elif not art_no and title:
        product = legal_name(title)
    else:
        raise ValueError("At least one of 'title' and 'art_no' must be provided")
    args = [brand, season, week, category, product, filename]
    return os.path.join(*map(legal_name, args))


def legal_name(name):
    """
    return a legal file name
    :param name: string
    :return: string

    >>> legal_name("?, /, \\, :, *, >, <, \'")
    '-, -, -, -, -, -, -, -'
    >>> legal_name('"')
    '-'

    """
    filename_char_set = set(name)
    illegal_chars = filename_char_set & ILLEGAL_FILENAME_CHARS
    for char in illegal_chars:
        name = name.replace(char, "-")
    return name


def gen_name_from_url(url):
    """
    generate a file name use url
    :param url: url without query
    :return: string of filename

    >>> gen_name_from_url('http://www.bearbrick.com/WI/upimage/0122_200724_pfhmgz_b.png')
    '0122_200724_pfhmgz_b.png'

    >>> gen_name_from_url('https://c.imgz.jp/802/53822802/53822802_46_d_500.jpg')
    '53822802_46_d_500.jpg'

    >>> gen_name_from_url('https://c.imgz.jp/802/53822802/53822802_46_d_500.jpg/')
    '53822802_46_d_500.jpg'
    """
    parts = url.split('/')
    if 'nike' in url:
        filename = unquote(parts[-2] + parts[-1])
    elif url.endswith('/'):
        filename = parts[-2]
    else:
        filename = parts[-1]
    return legal_name(filename)


def get_spider_cfg(name):
    """
    get parse item configuration of a specific spider.
    :param name: name of spider
    :return: dictionary of configuration
    """
    with open('products_crawler/spider_configs/{0}.json'.format(name), encoding='utf-8') as f:
        return json.load(f)



# -*- coding: utf-8 -*-

# Define here the models for your scraped items
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/items.html

from scrapy import Item, Field


class ProductItem(Item):
    brand = Field()
    title = Field()
    art_no = Field()
    item_url = Field()
    images = Field()
    price = Field()
    image_base_url = Field()
    season = Field()
    week = Field()
    category = Field()



# -*- coding: utf-8 -*-

# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html
import os.path
from os import makedirs
from scrapy import Request
from scrapy.pipelines.images import ImagesPipeline
from scrapy.exceptions import DropItem
from products_crawler.settings import IMAGES_STORE
from products_crawler.utils import file_path, gen_name_from_url


class ProductInfoPipeline(object):
    """save item info"""
    def process_item(self, item, spider):
        # 保存路径：
        # images/<brand>/<number>/url.txt
        filepath = file_path(item)
        filepath = os.path.join(IMAGES_STORE, filepath)
        try:
            makedirs(filepath)
        except OSError:
            pass
        with open(os.path.abspath(os.path.join(filepath, 'info.txt')), 'w', encoding='utf-8') as f:
            f.write(str(item.items()))
        return item


class ProductImagesPipeline(ImagesPipeline):
    def file_path(self, request, response=None, info=None, *, item=None):
        image_name = gen_name_from_url(request.url)
        file_name = file_path(request.meta['item'], image_name)
        return file_name

    def item_completed(self, results, item, info):
        image_paths = [x['path'] for ok, x in results if ok]
        if not image_paths:
            raise DropItem('Image Download Failed')
        return item

    def get_media_requests(self, item, info):
        image_base_url = item.get('image_base_url', None)
        if image_base_url:
            images_url = [image_base_url + part_url for part_url in item['images']]
        else:  # image_base_url is None
            images_url = item['images']
        for image_url in images_url:
            yield Request(image_url, meta={'item': item})



from selenium import webdriver
import time

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.chrome.options import Options

from scrapy.http import HtmlResponse
from logging import getLogger


class SeleniumDownloaderMiddleware(object):
    def __init__(self, timeout=None, chrome_options=None):
        self.logger = getLogger(__name__)
        self.timeout = timeout
        options = Options()
        for o in chrome_options or ():
            options.add_argument(o)
        self.browser = webdriver.Chrome(chrome_options=options)
        self.browser.set_window_size(1920, 1080)
        self.browser.set_page_load_timeout(timeout)
        self.wait = WebDriverWait(self.browser, self.timeout)
        self.actions = ActionChains(self.browser)

    def __del__(self):
        self.browser.quit()

    def process_request(self, request, spider):
        if request.url.endswith('jpg'):
            return None
        self.logger.debug('Chrome is starting')
        self.browser.get(request.url)
        self.actions.send_keys(Keys.END)
        self.actions.send_keys(Keys.END)
        self.actions.perform()
        time.sleep(1)
        return HtmlResponse(url=request.url, body=self.browser.page_source, request=request, status=200, encoding='utf-8')

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            timeout=crawler.settings.get('SELENIUM_TIMEOUT'),
            chrome_options=crawler.settings.get('CHROME_OPTIONS'),
        )



# -*- coding: utf-8 -*-
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Rule
from . import GenericSpider


class TsSpider(GenericSpider):
    name = 'ts'
    allowed_domains = ['travisscott.com']
    rules = (
        # Match each product in product list.
        Rule(LinkExtractor(allow=r'.*products.*'), callback='parse', follow=True),
    )



# -*- coding: utf-8 -*-
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Rule
from . import GenericSpider


class UastoreSpider(GenericSpider):
    name = 'uastore'
    allowed_domains = ['store.united-arrows.co.jp']

    rules = (
        # Match each product in product list.
        Rule(LinkExtractor(restrict_xpaths='//div[@id="itemList"]'), callback='parse'),
        # Match other pages.
        Rule(LinkExtractor(restrict_xpaths='//div[contains(@id, "pager")]'))
    )



# -*- coding: utf-8 -*-
from datetime import date
from scrapy import Request
from . import GenericSpider


class SupremeSpider(GenericSpider):
    name = 'supreme'
    allowed_domains = ['supremecommunity.com']
    details_base_url = 'https://www.supremecommunity.com/season/itemdetails/'
    image_base_url = 'https://www.supremecommunity.com'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.start_urls:
            today = date.today()
            year, week, _ = today.isocalendar()
            month = today.month
            season = f'spring-summer{year}' if 1 <= month <= 6 else f'fall-winter{year}'
            for day in range(1, 8):
                date_ = date.fromisocalendar(year, week, day)
                url = f'https://www.supremecommunity.com/season/{season}/droplist/{date_.isoformat()}/'
                self.start_urls.append(url)

    def parse_start_url(self, response, **kwargs):
        """
        分析当季的页面，生成每一周的请求对象。
        :param response: 由属性start_urls中url生成的响应对象。
        :return: 每一周的请求对象，回调方法为parse_week()。
        """
        # 得到每一周的相对地址
        weeks = response.xpath(
            r'//div[@class="catalog-inner"]//a/@href'
        ).getall()
        # 判断该页面是否应该直接交给parse_week()方法进行解析
        if weeks:
            for week in weeks:
                yield Request(
                    url=self.image_base_url+week,
                    callback=self.parse_week
                )
        else:
            yield Request(
                url=response.url,
                callback=self.parse_week
            )

    def parse_week(self, response):
        """
        分析每一周的页面，返回对每一个商品的请求。
        :param response: parse_start_url()方法中返回的请求对象对应的响应对象。
        :return: 对商品的请求对象，回调方法为parse_item()。
        """
        details = response.xpath('//div[@class="catalog-item-top"]/@data-itemid')
        for detail_no in details:
            yield Request(
                url=self.details_base_url+detail_no.get(),
                callback=self.parse,
            )
